@@ -1918,12 +1918,19 @@ static void AnsiToUTF8 (char *string, int len)
     size_t outsize = len - 1;
     int err = 0;
     char *old_locale;
+    iconv_t converter;
 
     memset(temp, 0, len);
     old_locale = setlocale (LC_CTYPE, "");
-    iconv_t converter = iconv_open ("UTF-8", "");
-    err = iconv (converter, &inp, &insize, &outp, &outsize);
-    iconv_close (converter);
+    converter = iconv_open ("UTF-8", "");
+
+    if (converter != (iconv_t) -1) {
+        err = iconv (converter, &inp, &insize, &outp, &outsize);
+        iconv_close (converter);
+    }
+    else
+        err = -1;
+
     setlocale (LC_CTYPE, old_locale);
 
     if (err == -1) {
