@@ -17,6 +17,12 @@
 #define FASTCALL
 #endif
 
+#if defined(WIN32) || \
+    (defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && (BYTE_ORDER == LITTLE_ENDIAN))
+#define BITSTREAM_SHORTS    // use "shorts" for reading/writing bitstreams
+                            //  (only works on little-endian machines)
+#endif
+
 #include <sys/types.h>
 
 // This header file contains all the definitions required by WavPack.
@@ -36,23 +42,31 @@ typedef float float32_t;
 #include <inttypes.h>
 #endif
 
-typedef unsigned char	uchar;
+typedef unsigned char   uchar;
 
 #if !defined(__GNUC__) || defined(WIN32)
-typedef unsigned short	ushort;
-typedef unsigned int	uint;
+typedef unsigned short  ushort;
+typedef unsigned int    uint;
 #endif
 
-// This structure is used to access the individual fields of 32-bit ieee
-// floating point numbers. This will not be compatible with compilers that
-// allocate bit fields from the most significant bits, although I'm not sure
-// how common that is.
+// Because the C99 specification states that "The order of allocation of
+// bit-ﬁelds within a unit (high-order to low-order or low-order to
+// high-order) is implementation-deﬁned" (6.7.2.1), I decided to change
+// the representation of floating-point values from a structure of
+// bit-fields to a 32-bit integer with access macros. Note that the WavPack
+// library doesn't use any floating-point math to implement compression of
+// floating-point data (although a little floating-point math is used in
+// high-level functions unrelated to the codec).
 
-typedef struct {
-    unsigned mantissa : 23;
-    unsigned exponent : 8;
-    unsigned sign : 1;
-} f32;
+typedef int32_t f32;
+
+#define get_mantissa(f)     ((f) & 0x7fffff)
+#define get_exponent(f)     (((f) >> 23) & 0xff)
+#define get_sign(f)         (((f) >> 31) & 0x1)
+
+#define set_mantissa(f,v)   (f) ^= (((f) ^ (v)) & 0x7fffff)
+#define set_exponent(f,v)   (f) ^= (((f) ^ ((v) << 23)) & 0x7f800000)
+#define set_sign(f,v)       (f) ^= (((f) ^ ((v) << 31)) & 0x80000000)
 
 #include <stdio.h>
 
@@ -78,7 +92,7 @@ typedef struct {
     int32_t tag_file_pos;
     ID3_Tag id3_tag;
     APE_Tag_Hdr ape_tag_hdr;
-    char *ape_tag_data;
+    unsigned char *ape_tag_data;
 } M_Tag;
 
 // RIFF / wav header formats (these occur at the beginning of both wav files
@@ -127,42 +141,42 @@ typedef struct {
 
 // or-values for "flags"
 
-#define BYTES_STORED	3	// 1-4 bytes/sample
-#define MONO_FLAG	4	// not stereo
-#define HYBRID_FLAG	8	// hybrid mode
-#define JOINT_STEREO	0x10	// joint stereo
-#define CROSS_DECORR	0x20	// no-delay cross decorrelation
-#define HYBRID_SHAPE	0x40	// noise shape (hybrid mode only)
-#define FLOAT_DATA	0x80	// ieee 32-bit floating point data
+#define BYTES_STORED    3       // 1-4 bytes/sample
+#define MONO_FLAG       4       // not stereo
+#define HYBRID_FLAG     8       // hybrid mode
+#define JOINT_STEREO    0x10    // joint stereo
+#define CROSS_DECORR    0x20    // no-delay cross decorrelation
+#define HYBRID_SHAPE    0x40    // noise shape (hybrid mode only)
+#define FLOAT_DATA      0x80    // ieee 32-bit floating point data
 
-#define INT32_DATA	0x100	// special extended int handling
-#define HYBRID_BITRATE	0x200	// bitrate noise (hybrid mode only)
-#define HYBRID_BALANCE	0x400	// balance noise (hybrid stereo mode only)
+#define INT32_DATA      0x100   // special extended int handling
+#define HYBRID_BITRATE  0x200   // bitrate noise (hybrid mode only)
+#define HYBRID_BALANCE  0x400   // balance noise (hybrid stereo mode only)
 
-#define INITIAL_BLOCK	0x800	// initial block of multichannel segment
-#define FINAL_BLOCK	0x1000	// final block of multichannel segment
+#define INITIAL_BLOCK   0x800   // initial block of multichannel segment
+#define FINAL_BLOCK     0x1000  // final block of multichannel segment
 
-#define SHIFT_LSB	13
-#define SHIFT_MASK	(0x1fL << SHIFT_LSB)
+#define SHIFT_LSB       13
+#define SHIFT_MASK      (0x1fL << SHIFT_LSB)
 
-#define MAG_LSB		18
-#define MAG_MASK	(0x1fL << MAG_LSB)
+#define MAG_LSB         18
+#define MAG_MASK        (0x1fL << MAG_LSB)
 
-#define SRATE_LSB	23
-#define SRATE_MASK	(0xfL << SRATE_LSB)
+#define SRATE_LSB       23
+#define SRATE_MASK      (0xfL << SRATE_LSB)
 
-#define FALSE_STEREO	0x40000000	// block is stereo, but data is mono
+#define FALSE_STEREO    0x40000000      // block is stereo, but data is mono
 
-#define IGNORED_FLAGS	0x18000000	// reserved, but ignore if encountered
-#define NEW_SHAPING	0x20000000	// use IIR filter for negative shaping
-#define UNKNOWN_FLAGS	0x80000000	// also reserved, but refuse decode if
-					//  encountered
+#define IGNORED_FLAGS   0x18000000      // reserved, but ignore if encountered
+#define NEW_SHAPING     0x20000000      // use IIR filter for negative shaping
+#define UNKNOWN_FLAGS   0x80000000      // also reserved, but refuse decode if
+                                        //  encountered
 
 #define MONO_DATA (MONO_FLAG | FALSE_STEREO)
 
-#define MIN_STREAM_VERS	    0x402	// lowest stream version we'll decode
-#define MAX_STREAM_VERS	    0x410	// highest stream version we'll decode or encode
-#define CUR_STREAM_VERS	    0x406	// stream version we are [normally] writing now
+#define MIN_STREAM_VERS     0x402       // lowest stream version we'll decode
+#define MAX_STREAM_VERS     0x410       // highest stream version we'll decode or encode
+#define CUR_STREAM_VERS     0x407       // stream version we are [normally] writing now
 
 
 //////////////////////////// WavPack Metadata /////////////////////////////////
@@ -175,33 +189,33 @@ typedef struct {
     uchar id;
 } WavpackMetadata;
 
-#define ID_UNIQUE		0x3f
-#define ID_OPTIONAL_DATA	0x20
-#define ID_ODD_SIZE		0x40
-#define ID_LARGE		0x80
+#define ID_UNIQUE               0x3f
+#define ID_OPTIONAL_DATA        0x20
+#define ID_ODD_SIZE             0x40
+#define ID_LARGE                0x80
 
-#define ID_DUMMY		0x0
-#define ID_ENCODER_INFO		0x1
-#define ID_DECORR_TERMS		0x2
-#define ID_DECORR_WEIGHTS	0x3
-#define ID_DECORR_SAMPLES	0x4
-#define ID_ENTROPY_VARS		0x5
-#define ID_HYBRID_PROFILE	0x6
-#define ID_SHAPING_WEIGHTS	0x7
-#define ID_FLOAT_INFO		0x8
-#define ID_INT32_INFO		0x9
-#define ID_WV_BITSTREAM		0xa
-#define ID_WVC_BITSTREAM	0xb
-#define ID_WVX_BITSTREAM	0xc
-#define ID_CHANNEL_INFO		0xd
+#define ID_DUMMY                0x0
+#define ID_ENCODER_INFO         0x1
+#define ID_DECORR_TERMS         0x2
+#define ID_DECORR_WEIGHTS       0x3
+#define ID_DECORR_SAMPLES       0x4
+#define ID_ENTROPY_VARS         0x5
+#define ID_HYBRID_PROFILE       0x6
+#define ID_SHAPING_WEIGHTS      0x7
+#define ID_FLOAT_INFO           0x8
+#define ID_INT32_INFO           0x9
+#define ID_WV_BITSTREAM         0xa
+#define ID_WVC_BITSTREAM        0xb
+#define ID_WVX_BITSTREAM        0xc
+#define ID_CHANNEL_INFO         0xd
 
-#define ID_RIFF_HEADER		(ID_OPTIONAL_DATA | 0x1)
-#define ID_RIFF_TRAILER		(ID_OPTIONAL_DATA | 0x2)
-#define ID_REPLAY_GAIN		(ID_OPTIONAL_DATA | 0x3)
-#define ID_CUESHEET		(ID_OPTIONAL_DATA | 0x4)
-#define ID_CONFIG_BLOCK		(ID_OPTIONAL_DATA | 0x5)
-#define ID_MD5_CHECKSUM		(ID_OPTIONAL_DATA | 0x6)
-#define ID_SAMPLE_RATE		(ID_OPTIONAL_DATA | 0x7)
+#define ID_RIFF_HEADER          (ID_OPTIONAL_DATA | 0x1)
+#define ID_RIFF_TRAILER         (ID_OPTIONAL_DATA | 0x2)
+#define ID_REPLAY_GAIN          (ID_OPTIONAL_DATA | 0x3)
+#define ID_CUESHEET             (ID_OPTIONAL_DATA | 0x4)
+#define ID_CONFIG_BLOCK         (ID_OPTIONAL_DATA | 0x5)
+#define ID_MD5_CHECKSUM         (ID_OPTIONAL_DATA | 0x6)
+#define ID_SAMPLE_RATE          (ID_OPTIONAL_DATA | 0x7)
 
 ///////////////////////// WavPack Configuration ///////////////////////////////
 
@@ -219,57 +233,63 @@ typedef struct {
     char **tag_strings;
 } WavpackConfig;
 
-#define CONFIG_BYTES_STORED	3	// 1-4 bytes/sample
-#define CONFIG_MONO_FLAG	4	// not stereo
-#define CONFIG_HYBRID_FLAG	8	// hybrid mode
-#define CONFIG_JOINT_STEREO	0x10	// joint stereo
-#define CONFIG_CROSS_DECORR	0x20	// no-delay cross decorrelation
-#define CONFIG_HYBRID_SHAPE	0x40	// noise shape (hybrid mode only)
-#define CONFIG_FLOAT_DATA	0x80	// ieee 32-bit floating point data
+#define CONFIG_BYTES_STORED     3       // 1-4 bytes/sample
+#define CONFIG_MONO_FLAG        4       // not stereo
+#define CONFIG_HYBRID_FLAG      8       // hybrid mode
+#define CONFIG_JOINT_STEREO     0x10    // joint stereo
+#define CONFIG_CROSS_DECORR     0x20    // no-delay cross decorrelation
+#define CONFIG_HYBRID_SHAPE     0x40    // noise shape (hybrid mode only)
+#define CONFIG_FLOAT_DATA       0x80    // ieee 32-bit floating point data
 
-#define CONFIG_FAST_FLAG	0x200	// fast mode
-#define CONFIG_HIGH_FLAG	0x800	// high quality mode
-#define CONFIG_VERY_HIGH_FLAG	0x1000	// very high
-#define CONFIG_BITRATE_KBPS	0x2000	// bitrate is kbps, not bits / sample
-#define CONFIG_AUTO_SHAPING	0x4000	// automatic noise shaping
-#define CONFIG_SHAPE_OVERRIDE	0x8000	// shaping mode specified
-#define CONFIG_JOINT_OVERRIDE	0x10000	// joint-stereo mode specified
-#define CONFIG_CREATE_EXE	0x40000	// create executable
-#define CONFIG_CREATE_WVC	0x80000	// create correction file
-#define CONFIG_OPTIMIZE_WVC	0x100000 // maximize bybrid compression
-#define CONFIG_CALC_NOISE	0x800000 // calc noise in hybrid mode
-#define CONFIG_LOSSY_MODE	0x1000000 // obsolete (for information)
-#define CONFIG_EXTRA_MODE	0x2000000 // extra processing mode
-#define CONFIG_SKIP_WVX		0x4000000 // no wvx stream w/ floats & big ints
-#define CONFIG_MD5_CHECKSUM	0x8000000 // compute & store MD5 signature
-#define CONFIG_OPTIMIZE_MONO	0x80000000 // optimize for mono streams posing as stereo
+#define CONFIG_FAST_FLAG        0x200   // fast mode
+#define CONFIG_HIGH_FLAG        0x800   // high quality mode
+#define CONFIG_VERY_HIGH_FLAG   0x1000  // very high
+#define CONFIG_BITRATE_KBPS     0x2000  // bitrate is kbps, not bits / sample
+#define CONFIG_AUTO_SHAPING     0x4000  // automatic noise shaping
+#define CONFIG_SHAPE_OVERRIDE   0x8000  // shaping mode specified
+#define CONFIG_JOINT_OVERRIDE   0x10000 // joint-stereo mode specified
+#define CONFIG_DYNAMIC_SHAPING  0x20000 // dynamic noise shaping
+#define CONFIG_CREATE_EXE       0x40000 // create executable
+#define CONFIG_CREATE_WVC       0x80000 // create correction file
+#define CONFIG_OPTIMIZE_WVC     0x100000 // maximize bybrid compression
+#define CONFIG_CALC_NOISE       0x800000 // calc noise in hybrid mode
+#define CONFIG_LOSSY_MODE       0x1000000 // obsolete (for information)
+#define CONFIG_EXTRA_MODE       0x2000000 // extra processing mode
+#define CONFIG_SKIP_WVX         0x4000000 // no wvx stream w/ floats & big ints
+#define CONFIG_MD5_CHECKSUM     0x8000000 // compute & store MD5 signature
+#define CONFIG_MERGE_BLOCKS     0x10000000 // merge blocks of equal redundancy (for lossyWAV)
+#define CONFIG_OPTIMIZE_MONO    0x80000000 // optimize for mono streams posing as stereo
 
 /*
- * These config flags are no longer used (or were never used) although there
- * may be WavPack files that have some of these bits set in the config
- * metadata structure, so be careful reusing them for something else.
+ * These config flags were never actually used, or are no longer used, or are
+ * used for something else now. They may be used in the future for what they
+ * say, or for something else. WavPack files in the wild *may* have some of
+ * these bit set in their config flags (with these older meanings), but only
+ * if the stream version is 0x410 or less than 0x407. Of course, this is not
+ * very important because once the file has been encoded, the config bits are
+ * just for information purposes (i.e., they do not affect decoding),
  *
-#define CONFIG_ADOBE_MODE	0x100	// "adobe" mode for 32-bit floats
-#define CONFIG_VERY_FAST_FLAG	0x400	// double fast
-#define CONFIG_COPY_TIME	0x20000	// copy file-time from source
-#define CONFIG_QUALITY_MODE	0x200000 // psychoacoustic quality mode
-#define CONFIG_RAW_FLAG		0x400000 // raw mode (not implemented yet)
-#define CONFIG_QUIET_MODE	0x10000000 // don't report progress %
-#define CONFIG_IGNORE_LENGTH	0x20000000 // ignore length in wav header
-#define CONFIG_NEW_RIFF_HEADER	0x40000000 // generate new RIFF wav header
+#define CONFIG_ADOBE_MODE       0x100   // "adobe" mode for 32-bit floats
+#define CONFIG_VERY_FAST_FLAG   0x400   // double fast
+#define CONFIG_COPY_TIME        0x20000 // copy file-time from source
+#define CONFIG_QUALITY_MODE     0x200000 // psychoacoustic quality mode
+#define CONFIG_RAW_FLAG         0x400000 // raw mode (not implemented yet)
+#define CONFIG_QUIET_MODE       0x10000000 // don't report progress %
+#define CONFIG_IGNORE_LENGTH    0x20000000 // ignore length in wav header
+#define CONFIG_NEW_RIFF_HEADER  0x40000000 // generate new RIFF wav header
  *
  */
 
-#define EXTRA_SCAN_ONLY		1
-#define EXTRA_STEREO_MODES	2
-#define EXTRA_TRY_DELTAS	8
-#define EXTRA_ADJUST_DELTAS	16
-#define EXTRA_SORT_FIRST	32
-#define EXTRA_BRANCHES		0x1c0
-#define EXTRA_SKIP_8TO16	512
-#define EXTRA_TERMS		0x3c00
-#define EXTRA_DUMP_TERMS	16384
-#define EXTRA_SORT_LAST		32768
+#define EXTRA_SCAN_ONLY         1
+#define EXTRA_STEREO_MODES      2
+#define EXTRA_TRY_DELTAS        8
+#define EXTRA_ADJUST_DELTAS     16
+#define EXTRA_SORT_FIRST        32
+#define EXTRA_BRANCHES          0x1c0
+#define EXTRA_SKIP_8TO16        512
+#define EXTRA_TERMS             0x3c00
+#define EXTRA_DUMP_TERMS        16384
+#define EXTRA_SORT_LAST         32768
 
 //////////////////////////////// WavPack Stream ///////////////////////////////
 
@@ -280,12 +300,17 @@ typedef struct {
 // possible to decode WavPack blocks without buffering an entire block.
 
 typedef struct bs {
-    uchar *buf, *end, *ptr;
+#ifdef BITSTREAM_SHORTS
+    unsigned short *buf, *end, *ptr;
+#else
+    unsigned char *buf, *end, *ptr;
+#endif
     void (*wrap)(struct bs *bs);
     int error, bc;
     uint32_t sr;
 } Bitstream;
 
+#define MAX_WRAPPER_BYTES 16777216
 #define MAX_STREAMS 8
 #define MAX_NTERMS 16
 #define MAX_TERM 8
@@ -301,8 +326,20 @@ typedef struct {
     char joint_stereo, delta, terms [MAX_NTERMS+1];
 } WavpackDecorrSpec;
 
+struct entropy_data {
+    uint32_t median [3], slow_level, error_limit;
+};
+
+struct words_data {
+    uint32_t bitrate_delta [2], bitrate_acc [2];
+    uint32_t pend_data, holding_one, zeros_acc;
+    int holding_zero, pend_count;
+    struct entropy_data c [2];
+};
+
 typedef struct {
     WavpackHeader wphdr;
+    struct words_data w;
 
     uchar *blockbuff, *blockend;
     uchar *block2buff, *block2end;
@@ -317,31 +354,26 @@ typedef struct {
 
     uchar int32_sent_bits, int32_zeros, int32_ones, int32_dups;
     uchar float_flags, float_shift, float_max_exp, float_norm_exp;
- 
+
     struct {
-	int32_t shaping_acc [2], shaping_delta [2], error [2];
-	double noise_sum, noise_ave, noise_max;
+        int32_t shaping_acc [2], shaping_delta [2], error [2];
+        double noise_sum, noise_ave, noise_max;
+        short *shaping_data, *shaping_array;
+        int32_t shaping_samples;
     } dc;
 
-    struct decorr_pass decorr_passes [MAX_NTERMS];
-    WavpackDecorrSpec *decorr_specs;
-
-    struct {
-	uint32_t bitrate_delta [2], bitrate_acc [2];
-	uint32_t median [3] [2], slow_level [2], error_limit [2];
-	uint32_t pend_data, holding_one, zeros_acc;
-	int holding_zero, pend_count;
-    } w;
+    struct decorr_pass decorr_passes [MAX_NTERMS], analysis_pass;
+    const WavpackDecorrSpec *decorr_specs;
 } WavpackStream;
 
 // flags for float_flags:
 
-#define FLOAT_SHIFT_ONES 1	// bits left-shifted into float = '1'
-#define FLOAT_SHIFT_SAME 2	// bits left-shifted into float are the same
-#define FLOAT_SHIFT_SENT 4	// bits shifted into float are sent literally
-#define FLOAT_ZEROS_SENT 8	// "zeros" are not all real zeros
-#define FLOAT_NEG_ZEROS  0x10	// contains negative zeros
-#define FLOAT_EXCEPTIONS 0x20	// contains exceptions (inf, nan, etc.)
+#define FLOAT_SHIFT_ONES 1      // bits left-shifted into float = '1'
+#define FLOAT_SHIFT_SAME 2      // bits left-shifted into float are the same
+#define FLOAT_SHIFT_SENT 4      // bits shifted into float are sent literally
+#define FLOAT_ZEROS_SENT 8      // "zeros" are not all real zeros
+#define FLOAT_NEG_ZEROS  0x10   // contains negative zeros
+#define FLOAT_EXCEPTIONS 0x20   // contains exceptions (inf, nan, etc.)
 
 /////////////////////////////// WavPack Context ///////////////////////////////
 
@@ -382,7 +414,7 @@ typedef struct {
 
     uint32_t filelen, file2len, filepos, file2pos, total_samples, crc_errors, first_flags;
     int wvc_flag, open_flags, norm_offset, reduced_channels, lossy_blocks, close_files;
-    uint32_t block_samples, max_samples, acc_samples, initial_index;
+    uint32_t block_samples, ave_block_samples, block_boundary, max_samples, acc_samples, initial_index;
     int riff_header_added, riff_header_created;
     M_Tag m_tag;
 
@@ -405,14 +437,14 @@ typedef struct {
 #define apply_weight_f(weight, sample) (((((sample & 0xffff) * weight) >> 9) + \
     (((sample & ~0xffff) >> 9) * weight) + 1) >> 1)
 
-#if 1	// PERFCOND
+#if 1   // PERFCOND
 #define apply_weight(weight, sample) (sample != (short) sample ? \
     apply_weight_f (weight, sample) : apply_weight_i (weight, sample))
 #else
 #define apply_weight(weight, sample) ((int32_t)((weight * (int64_t) sample + 512) >> 10))
 #endif
 
-#if 1	// PERFCOND
+#if 1   // PERFCOND
 #define update_weight(weight, delta, source, result) \
     if (source && result) { int32_t s = (int32_t) (source ^ result) >> 31; weight = (delta ^ s) + (weight - s); }
 #elif 1
@@ -423,28 +455,27 @@ typedef struct {
     if (source && result) (source ^ result) < 0 ? (weight -= delta) : (weight += delta);
 #endif
 
-#define update_weight_d1(weight, delta, source, result) \
-    if (source && result) weight += ((source ^ result) >> 30) | 1;
-
 #define update_weight_d2(weight, delta, source, result) \
     if (source && result) weight -= (((source ^ result) >> 29) & 4) - 2;
 
 #define update_weight_clip(weight, delta, source, result) \
-    if (source && result && ((source ^ result) < 0 ? (weight -= delta) < -1024 : (weight += delta) > 1024)) \
-	weight = weight < 0 ? -1024 : 1024;
-
-#define update_weight_clip_d1(weight, delta, source, result) \
-    if (source && result && abs (weight -= (((source ^ result) >> 30) & 2) - 1) > 1024) \
-	weight = weight < 0 ? -1024 : 1024;
+    if (source && result) { \
+        const int32_t s = (source ^ result) >> 31; \
+        if ((weight = (weight ^ s) + (delta - s)) > 1024) weight = 1024; \
+        weight = (weight ^ s) - s; \
+    }
 
 #define update_weight_clip_d2(weight, delta, source, result) \
-    if (source && result && abs (weight -= (((source ^ result) >> 29) & 4) - 2) > 1024) \
-	weight = weight < 0 ? -1024 : 1024;
+    if (source && result) { \
+        const int32_t s = (source ^ result) >> 31; \
+        if ((weight = (weight ^ s) + (2 - s)) > 1024) weight = 1024; \
+        weight = (weight ^ s) - s; \
+    }
 
 // bits.c
 
-void bs_open_read (Bitstream *bs, uchar *buffer_start, uchar *buffer_end);
-void bs_open_write (Bitstream *bs, uchar *buffer_start, uchar *buffer_end);
+void bs_open_read (Bitstream *bs, void *buffer_start, void *buffer_end);
+void bs_open_write (Bitstream *bs, void *buffer_start, void *buffer_end);
 uint32_t bs_close_read (Bitstream *bs);
 uint32_t bs_close_write (Bitstream *bs);
 
@@ -460,23 +491,23 @@ int DoCloseHandle (FILE *hFile), DoTruncateFile (FILE *hFile);
 
 #define getbit(bs) ( \
     (((bs)->bc) ? \
-	((bs)->bc--, (bs)->sr & 1) : \
-	    (((++((bs)->ptr) != (bs)->end) ? (void) 0 : (bs)->wrap (bs)), (bs)->bc = 7, ((bs)->sr = *((bs)->ptr)) & 1) \
+        ((bs)->bc--, (bs)->sr & 1) : \
+            (((++((bs)->ptr) != (bs)->end) ? (void) 0 : (bs)->wrap (bs)), (bs)->bc = sizeof (*((bs)->ptr)) * 8 - 1, ((bs)->sr = *((bs)->ptr)) & 1) \
     ) ? \
-	((bs)->sr >>= 1, 1) : \
-	((bs)->sr >>= 1, 0) \
+        ((bs)->sr >>= 1, 1) : \
+        ((bs)->sr >>= 1, 0) \
 )
 
 #define getbits(value, nbits, bs) { \
     while ((nbits) > (bs)->bc) { \
-	if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
-	(bs)->sr |= (int32_t)*((bs)->ptr) << (bs)->bc; \
-	(bs)->bc += 8; \
+        if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
+        (bs)->sr |= (int32_t)*((bs)->ptr) << (bs)->bc; \
+        (bs)->bc += sizeof (*((bs)->ptr)) * 8; \
     } \
     *(value) = (bs)->sr; \
     if ((bs)->bc > 32) { \
         (bs)->bc -= (nbits); \
-        (bs)->sr = *((bs)->ptr) >> (8 - (bs)->bc); \
+        (bs)->sr = *((bs)->ptr) >> (sizeof (*((bs)->ptr)) * 8 - (bs)->bc); \
     } \
     else { \
         (bs)->bc -= (nbits); \
@@ -485,35 +516,36 @@ int DoCloseHandle (FILE *hFile), DoTruncateFile (FILE *hFile);
 }
 
 #define putbit(bit, bs) { if (bit) (bs)->sr |= (1 << (bs)->bc); \
-    if (++((bs)->bc) == 8) { \
-	*((bs)->ptr) = (bs)->sr; \
-	(bs)->sr = (bs)->bc = 0; \
-	if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
+    if (++((bs)->bc) == sizeof (*((bs)->ptr)) * 8) { \
+        *((bs)->ptr) = (bs)->sr; \
+        (bs)->sr = (bs)->bc = 0; \
+        if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
     }}
 
 #define putbit_0(bs) { \
-    if (++((bs)->bc) == 8) { \
-	*((bs)->ptr) = (bs)->sr; \
-	(bs)->sr = (bs)->bc = 0; \
-	if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
+    if (++((bs)->bc) == sizeof (*((bs)->ptr)) * 8) { \
+        *((bs)->ptr) = (bs)->sr; \
+        (bs)->sr = (bs)->bc = 0; \
+        if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
     }}
 
 #define putbit_1(bs) { (bs)->sr |= (1 << (bs)->bc); \
-    if (++((bs)->bc) == 8) { \
-	*((bs)->ptr) = (bs)->sr; \
-	(bs)->sr = (bs)->bc = 0; \
-	if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
+    if (++((bs)->bc) == sizeof (*((bs)->ptr)) * 8) { \
+        *((bs)->ptr) = (bs)->sr; \
+        (bs)->sr = (bs)->bc = 0; \
+        if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
     }}
 
 #define putbits(value, nbits, bs) { \
     (bs)->sr |= (int32_t)(value) << (bs)->bc; \
-    if (((bs)->bc += (nbits)) >= 8) \
-	do { \
-	    *((bs)->ptr) = (bs)->sr; \
-	    (bs)->sr >>= 8; \
-	    if (((bs)->bc -= 8) > 24) (bs)->sr |= ((value) >> ((nbits) - (bs)->bc)); \
-	    if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
-	} while ((bs)->bc >= 8); \
+    if (((bs)->bc += (nbits)) >= sizeof (*((bs)->ptr)) * 8) \
+        do { \
+            *((bs)->ptr) = (bs)->sr; \
+            (bs)->sr >>= sizeof (*((bs)->ptr)) * 8; \
+            if (((bs)->bc -= sizeof (*((bs)->ptr)) * 8) > 32 - sizeof (*((bs)->ptr)) * 8) \
+                (bs)->sr |= ((value) >> ((nbits) - (bs)->bc)); \
+            if (++((bs)->ptr) == (bs)->end) (bs)->wrap (bs); \
+        } while ((bs)->bc >= sizeof (*((bs)->ptr)) * 8); \
 }
 
 void little_endian_to_native (void *data, char *format);
@@ -571,9 +603,9 @@ void write_hybrid_profile (WavpackStream *wps, WavpackMetadata *wpmd);
 int read_entropy_vars (WavpackStream *wps, WavpackMetadata *wpmd);
 int read_hybrid_profile (WavpackStream *wps, WavpackMetadata *wpmd);
 int32_t FASTCALL send_word (WavpackStream *wps, int32_t value, int chan);
-void FASTCALL send_word_lossless (WavpackStream *wps, int32_t value, int chan);
+void send_words_lossless (WavpackStream *wps, int32_t *buffer, int32_t nsamples);
 int32_t FASTCALL get_word (WavpackStream *wps, int chan, int32_t *correction);
-int32_t FASTCALL get_word_lossless (WavpackStream *wps, int chan);
+int32_t get_words_lossless (WavpackStream *wps, int32_t *buffer, int32_t nsamples);
 void flush_word (WavpackStream *wps);
 int32_t nosend_word (WavpackStream *wps, int32_t value, int chan);
 void scan_word (WavpackStream *wps, int32_t *samples, uint32_t num_samples, int dir);
@@ -585,7 +617,7 @@ uint32_t log2buffer (int32_t *samples, uint32_t num_samples, int limit);
 signed char store_weight (int weight);
 int restore_weight (signed char weight);
 
-#define WORD_EOF (1L << 31)
+#define WORD_EOF ((int32_t)(1L << 31))
 
 // float.c
 
@@ -608,30 +640,33 @@ void execute_mono (WavpackContext *wpc, int32_t *samples, int no_history, int do
 WavpackContext *WavpackOpenFileInputEx (WavpackStreamReader *reader, void *wv_id, void *wvc_id, char *error, int flags, int norm_offset);
 WavpackContext *WavpackOpenFileInput (const char *infilename, char *error, int flags, int norm_offset);
 
-#define OPEN_WVC	0x1	// open/read "correction" file
-#define OPEN_TAGS	0x2	// read ID3v1 / APEv2 tags (seekable file)
-#define OPEN_WRAPPER	0x4	// make audio wrapper available (i.e. RIFF)
-#define OPEN_2CH_MAX	0x8	// open multichannel as stereo (no downmix)
-#define OPEN_NORMALIZE	0x10	// normalize floating point data to +/- 1.0
-#define OPEN_STREAMING	0x20	// "streaming" mode blindly unpacks blocks
-				// w/o regard to header file position info
-#define OPEN_EDIT_TAGS	0x40	// allow editing of tags
+#define OPEN_WVC        0x1     // open/read "correction" file
+#define OPEN_TAGS       0x2     // read ID3v1 / APEv2 tags (seekable file)
+#define OPEN_WRAPPER    0x4     // make audio wrapper available (i.e. RIFF)
+#define OPEN_2CH_MAX    0x8     // open multichannel as stereo (no downmix)
+#define OPEN_NORMALIZE  0x10    // normalize floating point data to +/- 1.0
+#define OPEN_STREAMING  0x20    // "streaming" mode blindly unpacks blocks
+                                // w/o regard to header file position info
+#define OPEN_EDIT_TAGS  0x40    // allow editing of tags
 
 int WavpackGetMode (WavpackContext *wpc);
 
-#define MODE_WVC	0x1
-#define MODE_LOSSLESS	0x2
-#define MODE_HYBRID	0x4
-#define MODE_FLOAT	0x8
-#define MODE_VALID_TAG	0x10
-#define MODE_HIGH	0x20
-#define MODE_FAST	0x40
-#define MODE_EXTRA	0x80
-#define MODE_APETAG	0x100
-#define MODE_SFX	0x200
-#define MODE_VERY_HIGH	0x400
-#define MODE_MD5	0x800
+#define MODE_WVC        0x1
+#define MODE_LOSSLESS   0x2
+#define MODE_HYBRID     0x4
+#define MODE_FLOAT      0x8
+#define MODE_VALID_TAG  0x10
+#define MODE_HIGH       0x20
+#define MODE_FAST       0x40
+#define MODE_EXTRA      0x80    // extra mode used, see MODE_XMODE for possible level
+#define MODE_APETAG     0x100
+#define MODE_SFX        0x200
+#define MODE_VERY_HIGH  0x400
+#define MODE_MD5        0x800
+#define MODE_XMODE      0x7000  // mask for extra level (1-6, 0=unknown)
+#define MODE_DNS        0x8000
 
+char *WavpackGetErrorMessage (WavpackContext *wpc);
 int WavpackGetVersion (WavpackContext *wpc);
 uint32_t WavpackUnpackSamples (WavpackContext *wpc, int32_t *buffer, uint32_t samples);
 uint32_t WavpackGetNumSamples (WavpackContext *wpc);
@@ -646,6 +681,7 @@ int WavpackGetBytesPerSample (WavpackContext *wpc);
 int WavpackGetNumChannels (WavpackContext *wpc);
 int WavpackGetChannelMask (WavpackContext *wpc);
 int WavpackGetReducedChannels (WavpackContext *wpc);
+int WavpackGetFloatNormExp (WavpackContext *wpc);
 int WavpackGetMD5Sum (WavpackContext *wpc, uchar data [16]);
 uint32_t WavpackGetWrapperBytes (WavpackContext *wpc);
 uchar *WavpackGetWrapperData (WavpackContext *wpc);
@@ -676,4 +712,53 @@ void *WavpackGetWrapperLocation (void *first_block, uint32_t *size);
 void WavpackLittleEndianToNative (void *data, char *format);
 void WavpackNativeToLittleEndian (void *data, char *format);
 
+uint32_t WavpackGetLibraryVersion (void);
+const char *WavpackGetLibraryVersionString (void);
+
+///////////////////////////// SIMD helper macros /////////////////////////////
+
+#ifdef OPT_MMX
+
+#if defined (__GNUC__) && !defined (__INTEL_COMPILER)
+//directly map to gcc's native builtins for faster code
+
+#if __GNUC__ < 4
+typedef int __di __attribute__ ((__mode__ (__DI__)));
+typedef int __m64 __attribute__ ((__mode__ (__V2SI__)));
+typedef int __v4hi __attribute__ ((__mode__ (__V4HI__)));
+#define _m_paddsw(m1, m2) (__m64) __builtin_ia32_paddsw ((__v4hi) m1, (__v4hi) m2)
+#define _m_pand(m1, m2) (__m64) __builtin_ia32_pand ((__di) m1, (__di) m2)
+#define _m_pandn(m1, m2) (__m64) __builtin_ia32_pandn ((__di) m1, (__di) m2)
+#define _m_pmaddwd(m1, m2) __builtin_ia32_pmaddwd ((__v4hi) m1, (__v4hi) m2)
+#define _m_por(m1, m2) (__m64) __builtin_ia32_por ((__di) m1, (__di) m2)
+#define _m_pxor(m1, m2) (__m64) __builtin_ia32_pxor ((__di) m1, (__di) m2)
+#else
+typedef int __m64 __attribute__ ((__vector_size__ (8)));
+typedef short __m64_16 __attribute__ ((__vector_size__ (8)));
+#define _m_paddsw(m1, m2) (__m64) __builtin_ia32_paddsw ((__m64_16) m1, (__m64_16) m2)
+#define _m_pand(m1, m2) __builtin_ia32_pand (m1, m2)
+#define _m_pandn(m1, m2) __builtin_ia32_pandn (m1, m2)
+#define _m_pmaddwd(m1, m2) __builtin_ia32_pmaddwd ((__m64_16) m1, (__m64_16) m2)
+#define _m_por(m1, m2) __builtin_ia32_por (m1, m2)
+#define _m_pxor(m1, m2) __builtin_ia32_pxor (m1, m2)
+#endif
+
+#define _m_paddd(m1, m2) __builtin_ia32_paddd (m1, m2)
+#define _m_pcmpeqd(m1, m2) __builtin_ia32_pcmpeqd (m1, m2)
+#define _m_pslldi(m1, m2) __builtin_ia32_pslld (m1, m2)
+#define _m_psradi(m1, m2) __builtin_ia32_psrad (m1, m2)
+#define _m_psrldi(m1, m2) __builtin_ia32_psrld (m1, m2)
+#define _m_psubd(m1, m2) __builtin_ia32_psubd (m1, m2)
+#define _m_punpckhdq(m1, m2) __builtin_ia32_punpckhdq (m1, m2)
+#define _m_punpckldq(m1, m2) __builtin_ia32_punpckldq (m1, m2)
+#define _mm_empty() __builtin_ia32_emms ()
+#define _mm_set_pi32(m1, m2) { m2, m1 }
+#define _mm_set1_pi32(m) { m, m }
+
+#else
+#include <mmintrin.h>
+#endif
+
+#endif //OPT_MMX
+
 #endif
